#import "PNNet.h"

#import "PNPlace.h"
#import "PNTransition.h"

NSString* PNNetWillChangeMarkings = @"PNNetWillChangeMarkings";
NSString* PNNetDidChangeMarkings = @"PNNetDidChangeMarkings";


@interface PNNet ()
@property (readonly, strong) NSOperationQueue *sequenceQueue;
@end

@implementation PNNet

@dynamic sequenceQueue;

// Custom logic goes here.

-(void)lazilyPostOnMainThread:(NSString*)notification
{
    NSRunLoop *mainRunLoop = [NSRunLoop mainRunLoop]; // [NSRunLoop currentRunLoop]
    [mainRunLoop cancelPerformSelector: @selector(lazilyPostOnMainThread:)
                                target: self
                              argument: notification];
    
    NSNotificationCenter *dfc = [NSNotificationCenter defaultCenter];
    [dfc postNotificationName: notification
                       object: self];
}
-(void)postOnMainThread: (NSString*)notification
{
    /*
     performSelector:target:argument:order:modes:
     
     Discussion
     This method sets up a timer to perform the aSelector message
     on the current thread’s run loop
     at the start of the next run loop iteration.
     */
    NSArray *modes = [NSArray arrayWithObject: NSDefaultRunLoopMode];
    NSRunLoop *mainRunLoop = [NSRunLoop mainRunLoop];
    [mainRunLoop performSelector: @selector(lazilyPostOnMainThread:)
                          target: self
                        argument: notification
                           order: 5
                           modes: modes];
}


#pragma mark These methods provide unique names for places and transitions
+(NSString*)placeNameWithIndex: (NSUInteger)index
{
    return [NSString stringWithFormat: @"P_%d", index];
}
+(NSString*)transitionNameWithIndex: (NSUInteger)index
{
    return [NSString stringWithFormat: @"T_%d", index];
}
-(NSSet*)allPlaceNames
{
    NSMutableSet *allPlaceNames = [NSMutableSet setWithCapacity: self.places.count];
    
    for( PNPlace *next in self.places ){
        [allPlaceNames addObject: next.name];
    }
    return allPlaceNames;
}
-(NSSet*)allTransitionNames
{
    NSMutableSet *allTransitionNames = [NSMutableSet setWithCapacity: self.transitions.count];
    
    for( PNTransition *next in self.transitions ){
        [allTransitionNames addObject: next.name];
    }
    return allTransitionNames;
}

-(NSString*)nextPlaceName
{
    NSUInteger nextIndex = 1;
    NSString *nextPlaceName = [PNNet placeNameWithIndex: nextIndex];
    NSSet *allPlaceNames = [self allPlaceNames];
    
    while( [allPlaceNames containsObject: nextPlaceName] ){
        nextPlaceName = [PNNet placeNameWithIndex: ++nextIndex];
    }
    return nextPlaceName;
}
-(NSString*)nextTransitionName
{
    NSUInteger nextIndex = 1;
    NSString *nextTransitionName = [PNNet transitionNameWithIndex: nextIndex];
    NSSet *allTransitionNames = [self allTransitionNames];
    
    while( [allTransitionNames containsObject: nextTransitionName] ){
        nextTransitionName = [PNNet transitionNameWithIndex: ++nextIndex];
    }
    return nextTransitionName;
}

#pragma mark override add/remove accessors to provide unique names for places and transitions
- (void)addPlaces:(NSSet*)morePlaces
{
    for( PNPlace *next in morePlaces ){
        [self addPlacesObject: next];
    }
}

- (void)addPlace:(PNPlace*)aPlace
{
    if( 0 == aPlace.name.length ){
        aPlace.name = [self nextPlaceName];
    }
    if( [[self allPlaceNames] containsObject: aPlace.name] ){
        aPlace.name = [self nextPlaceName];
    }
    
    [self addPlacesObject: aPlace];
}

- (void)addTransitions:(NSSet*)moreTransitions
{
    for( PNTransition *next in moreTransitions ){
        [self addTransition: next];
    }
}
- (void)addTransition:(PNTransition*)aTransition
{
    if( 0 == aTransition.name.length ){
        aTransition.name = [self nextTransitionName];
    }
    if( [[self allTransitionNames] containsObject: aTransition.name] ){
        aTransition.name = [self nextTransitionName];
    }

    [self addTransitionsObject: aTransition];
}


#pragma mark Accessing components by name:

# pragma mark two dictionaries
-(NSDictionary*)mapNamesToPlaces
{
    NSMutableDictionary *result =
    [NSMutableDictionary dictionaryWithCapacity: self.places.count];
    
    for( PNPlace *next in self.places ){
        [result setObject: next
                   forKey: next.name];
    }
    return result;
}
-(NSDictionary*)mapNamesToTransitions
{
    NSMutableDictionary *result =
    [NSMutableDictionary dictionaryWithCapacity: self.transitions.count];
    
    for( PNTransition *next in self.transitions ){
        [result setObject: next
                   forKey: next.name];
    }
    return result;
}
# pragma mark two accessors for nodes by name:

-(PNPlace*)placeForName: (NSString*)placeName
{
    NSDictionary *map = [self mapNamesToPlaces];
    return [map objectForKey: placeName];
}
-(PNTransition*)transitionForName: (NSString*)transitionName
{
    NSDictionary *map = [self mapNamesToTransitions];
    return [map objectForKey: transitionName];
}


#pragma mark Methods implementing the firing sequence:
-(NSOperationQueue*)sequenceQueue
{
    if( nil == m_sequenceQueue ){
        m_sequenceQueue = [[NSOperationQueue alloc] init];
        m_sequenceQueue.maxConcurrentOperationCount = 1;
    }
    return m_sequenceQueue;
}
-(void)addToFiringSequence: (PNTransition*)aTransition
{
    NSAssert( [self.transitions containsObject: aTransition],
             @"Bogus transition, not in this net.");
    
    BOOL savedState = [self.sequenceQueue isSuspended];
    
    [self.sequenceQueue setSuspended: YES];{ // "Suspending a queue does not stop operations that are already running."
                
        NSInvocationOperation *next =
        [[NSInvocationOperation alloc] initWithTarget:aTransition
                                             selector:@selector(fire)
                                               object:nil];
        
        NSUInteger count = self.sequenceQueue.operationCount;
        if( 0 != count ){
            /*
             
             Additional Operation Queue Behaviors
             An operation queue executes its queued operation objects based on their priority and readiness.
             If all of the queued operation objects have the same priority and are ready to execute
             when they are put in the queue
             that is, their isReady method returns YES
             they are executed in the order in which they were submitted to the queue.
             For a queue whose maximum number of concurrent operations is set to 1,
             this equates to a serial queue.
             
             However, you should never rely on the serial execution of operation objects.
             Changes in the readiness of an operation can change the resulting execution order.
             
             */
            NSOperation *previous =
            [[self.sequenceQueue operations] objectAtIndex: count-1];
            
            [next addDependency: previous];
        }
        
        [self.sequenceQueue addOperation: next];
        
    }[self.sequenceQueue setSuspended: savedState];
}
-(void)willFireTransition: (PNTransition*)aTransition
{
    [self.sequenceQueue setSuspended: YES];
    [self postOnMainThread: PNNetWillChangeMarkings];
}
-(void)didFireTransition: (PNTransition*)aTransition
            successfully: (BOOL)success
{
    [self.sequenceQueue setSuspended: !success];
    [self postOnMainThread: PNNetDidChangeMarkings];
}
@end
