//
//  PNDocToken+FileOperations.m
//  TaggedLocationsDemo
//
//  Created by Don Briggs on 6/22/12.
//  Copyright (c) 2012 Directed Graph Software LLC (AB). All rights reserved.
//

#import "PNDocToken+FileOperations.h"

#import "UIManagedDocument+PNDocTokenCallbackSupporting.h"
#import "PNDocToken+CallbackSupporting.h"
#import "PNDocToken+Managing.h"

#import "PNCloudURLToken.h"
#import "PNCloudURLToken+iCloudTransferring.h"

#import "NSURL+NPAssisting.h"

#import "PNURLToken.h"
#import "PNURLToken+LifeCycling.h"

#import "PNDocTokenPlace.h"

#import "PNDocTokenMarking.h"

#import "NSManagedObject+PNPermanentIDing.h"

#import "PNDocSupportNetController.h"
#import "PNDocTokenTransition.h"


@interface PNDocToken (OperationPipelining)

// Enqueue the selector (of a primary operation) in the operation queue.
// The operation queue performs one operation at a time.
// Use this method repeatedly to form compound operations.
-(void)enqueueSelector: (SEL)selectorArg;
-(void)redact;
-(void)restore;
-(void)openQueue;
-(void)closeQueue;

-(void)becomeAvailable;

@end


@implementation PNDocToken (OperationPipelining)

-(void)enqueueSelector: (SEL)selectorArg
{    
    NSLog(@"-[PNDocToken enqueueSelector: %@]",
          NSStringFromSelector( selectorArg ) );
    
    NSAssert( [self respondsToSelector: selectorArg], @"invalid selector");
    
    NSInvocationOperation *next = 
    [[NSInvocationOperation alloc] initWithTarget:self
                                         selector:selectorArg
                                           object:nil];
    
    [self.operationQueue setSuspended: YES];
    @synchronized(self){
        
        NSUInteger count = self.operationQueue.operationCount;
        if( 0 != count ){
            /*
             
             Additional Operation Queue Behaviors
             An operation queue executes its queued operation objects based on their priority and readiness.
             If all of the queued operation objects have the same priority and are ready to execute 
             when they are put in the queue
             that is, their isReady method returns YES
             they are executed in the order in which they were submitted to the queue. 
             For a queue whose maximum number of concurrent operations is set to 1, 
             this equates to a serial queue. 
             
             However, you should never rely on the serial execution of operation objects. 
             Changes in the readiness of an operation can change the resulting execution order.
             
             */
            NSOperation *previous = 
            [[self.operationQueue operations] objectAtIndex: count-1];
            [next addDependency: previous];
        }
        [self.operationQueue addOperation: next];

    } // @synchronized (self)
}
-(void)resetAndSuspendQueue
{
    [self.operationQueue setSuspended: NO];{
        NSUInteger opQCount = self.operationQueue.operationCount;
        NSLog(@"opQCount = %d", opQCount);
        [self.operationQueue cancelAllOperations];
        m_operationQueue = nil;
        
    }[self.operationQueue setSuspended: YES];
    // Using lazy instantiation for the operationQueue.
    // If ( nil == m_operationQueue ),
    // a call to self.operationQueue generates a new instance.
    
    [m_queueLog setString:@""];
}
-(void)redact // meaning: "bring back", "take back"
{
    NSLog(@"-[PNDocToken redact]");
    self.onceAndFuturePlace = (PNDocTokenPlace*)self.place;
    NSLog(@" self.place.marker = %@", self.onceAndFuturePlace.marker);
    
    // First operation in the queue
    [self.place removeToken: self];
}
-(void)restore
{
    NSLog(@"-[PNDocToken restore]");
    
    // Next to last operation in the queue (before -becomeAvailable).
    [self.onceAndFuturePlace addToken: self];
    NSLog(@" self.place.marker = %@", self.onceAndFuturePlace.marker);
}
    
-(void)openQueue
{
    [self resetAndSuspendQueue];
    [self.operationQueue setSuspended: YES];
    [self enqueueSelector:@selector(redact)];
}
-(void)closeQueue
{
    [self enqueueSelector:@selector(restore)];
    [self enqueueSelector:@selector(becomeAvailable)];
    [self.operationQueue setSuspended: NO];
}
-(void)becomeAvailable
{
    //NSAssert( !self.isAvailableValue, @"must be unavailable");
    
    // Don't call self.setIsAvailableValue:
    // Use the related method to avoid an infinite loop.
    self.isAvailableValue = YES;
    
    dispatch_async(dispatch_get_main_queue(),
                   ^{
                       PNDocTokenPlace *p = (PNDocTokenPlace*)(self.onceAndFuturePlace);
                       NSObject *m = (NSObject *)(p.marker);
                       
                       NSLog(@"self.place.marker = %@",
                             [m description]);
                       //[((PNDocTokenPlace*)(self.place)).marker docTokenDidBecomeAvailable: self];
                       id <PNDocTokenMarking> mPrime = (id <PNDocTokenMarking>)m;
                       [mPrime docTokenDidBecomeAvailable: self];
                   }
                   );
}



@end

@implementation PNDocToken (SimpleFileOperations)

/*
 A design decision requires that all doc token file operations 
 be performed by the operation queue.
 
 Each externally available method in the category serves as a wrapper. 
 Each such method adds to the operation queue a related method that performs the intended work.
 
 The pattern for simple file operations is:
 
 -(void)nqd_foo
 {
    // -nqd_foo is NOT declared in PNDocToken+FileOperations.h
    // It's enqueued in -foo, below.
 
    // Performs amazing UIManagedDocument foo...
 
    // If the enqueued method invokes a callback file operation,
    // i.e. invokes a method of the 
    // UIManagedDocument+PNDocTokenCallbackSupporting category,
    // it must first suspend the operation queue, 
    // then call the callback-supporting method:
 
    [self.operationQueue setSuspended: YES]; // suspend the queue
    [self.document closeWithToken: self];  // -closeWithToken: results in a callback.
 
 }
 -(void)foo
 {
    // -foo is declared in PNDocToken+FileOperations.h
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_foo)];
    }[self closeQueue];

 }
  
 The pattern for aggregate file operations is:
 Every -enqueueSelector: call must be nested in a an -openQueue/-closeQueue brace:
 
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_sel_1)];
        [self enqueueSelector: @selector(nqd_sel_2)];
        ...
        [self enqueueSelector: @selector(nqd_sel_n)];
    }[self closeQueue];
 

 */

-(void)nqd_allocInitDocumentWithLocalURL
{
    [self.operationQueue setSuspended: YES];

    NSLog(@"-[PNDocToken nqd_allocInitDocumentWithLocalURL] (%@) ",
          self.urlToken.localizedName);

    NSString *documentClassName = self.urlToken.documentClassName;
    if( nil == documentClassName && (nil != self.creationDelegate)){
        documentClassName = [self.creationDelegate documentClassNameForUrlToken: self.urlToken];
        self.urlToken.documentClassName = documentClassName;
    }
    Class docFactory = NSClassFromString(self.urlToken.documentClassName);
    
    NSAssert( [docFactory isSubclassOfClass: [UIManagedDocument class]], 
             @"%@ is not a subclass of UIManagedDocument.", 
             documentClassName);
    
    m_document = nil;
    m_document = [[docFactory alloc] initWithFileURL: self.urlToken.localFilePackageURL];
    
    self.document.persistentStoreOptions = self.urlToken.persistentStoreOptions;
    
    [self.operationQueue setSuspended: NO];

}
-(void)nqd_configureNewDocument
{
    [self.operationQueue setSuspended: YES];

    NSLog(@"-[PNDocToken nqd_configureNewDocument] (%@) ",
          self.urlToken.localizedName);

    if( NO == m_documentObjectGraphInitialized ){
        [self.creationDelegate initializeDocumentObjectGraphInDocToken: self];
        m_documentObjectGraphInitialized = YES;
        
        NSManagedObjectContext *moc = self.document.managedObjectContext;
        [moc processPendingChanges];
    }
    
    [self.operationQueue setSuspended: NO];

}
-(void)nqd_makeUbiquitous
{
    [self.operationQueue setSuspended: YES];
    NSLog(@"-[PNDocToken nqd_makeUbiquitous] (%@) ",
          self.urlToken.localizedName);

    NSFileManager *fMgr = [PNDocSupportNetController threadSafeFileManager];
    BOOL alreadyUbiquitous = [fMgr isUbiquitousItemAtURL: self.urlToken.localFilePackageURL];
    BOOL success = NO;
    NSError *localError = nil;
    
    if(alreadyUbiquitous){
        success = YES;
    }else{
        /**
         If we initialized and saved (saved-for-creating) and closed the doc with
         the cloud version of persistent store options,
         the cloud-based filePackageURL should exist already:
         
         Or not...
         */
        
        NSAssert( [self.urlToken isKindOfClass: [PNCloudURLToken class]],
                 @"Bogus url token.");
        
        PNCloudURLToken *cloudToken = (PNCloudURLToken*)self.urlToken;
        
        success = [cloudToken setUbiquitous: YES
                                      error:&localError];
        
        NSLog(@"-[PNDocToken nqd_makeUbiquitous] success = %@",
              success ? @"YES" : @"NO");
        
    }
    
    [self handleSuccess: success
              operation: _cmd
                  error: localError];
}
-(void)nqd_saveDocumentLocallyForCreating
{
    [self.operationQueue setSuspended: YES];
    
    NSLog(@"-[PNDocToken nqd_saveDocumentLocallyForCreating] (%@) ",
          self.urlToken.localizedName);
    
    enum UIDocumentState test = self.document.documentState;
    BOOL docStateOK =
    (UIDocumentStateClosed == test) || // Typical case
    (UIDocumentStateNormal == test);  // If called from PNRenameDoc
    
    NSAssert( docStateOK,  // If called from PNRenameDoc
             @"Document must be in its closed or normal state.");
    // Can the document be in conflict?
    
    //[self logRegisteredObjects];
    
    
    // If:
    // (a) the token is NPCloudURLToken; and
    // (b) the save operation is UIDocumentSaveForCreating; 
    // Then:
    //     the callback sets the local URL to be ubiquitous at the cloud URL.
    [self.document saveToURL: self.urlToken.localFilePackageURL
            forSaveOperation: UIDocumentSaveForCreating
                       token: self];
}
-(void)saveDocumentForCreating
{
    NSLog(@"-[PNDocToken saveDocumentForCreating]");
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_saveDocumentLocallyForCreating)];
    }[self closeQueue];
}
-(void)nqd_copyDocument
{
    [self.operationQueue setSuspended: YES];

    NSLog(@"-[PNDocToken nqd_copyDocument] (%@) ",
          self.urlToken.localizedName);
    
    NSAssert( 0 != (UIDocumentStateClosed & self.document.documentState),
             @"Document must be in its closed state.");
    NSAssert( (UIDocumentStateClosed == self.document.documentState),
             @"Document must be in its closed state exactly.");
    NSAssert( (nil == m_error),
             @"Existing error, cannot proceed: %@.",
             [m_error description] );
   
    NSError __block *localError = nil;
    
    BOOL __block success = NO;
    
    NSFileManager __block *fMgr = [PNDocSupportNetController threadSafeFileManager];
    
    NSFileCoordinator* coordinator =
    [[NSFileCoordinator alloc] initWithFilePresenter:self.document];
    [coordinator coordinateReadingItemAtURL: self.document.fileURL
                                    options: NSFileCoordinatorReadingWithoutChanges
                           writingItemAtURL: self.urlToken.localFilePackageURL
                                    options: NSFileCoordinatorWritingForMoving
                                      error: &localError
                                 byAccessor: ^(NSURL *newReadingURL, NSURL *newWritingURL)
     {
         success = [fMgr copyItemAtURL: newReadingURL
                                 toURL: newWritingURL
                                 error: &localError];
     }];

    [self handleSuccess: success
              operation: _cmd
                  error: localError];
    
    //NSAssert( success, @"Failed to copy");
    
}
-(void)nqd_openDocument
{
    NSLog(@"-[PNDocToken nqd_openDocument] (%@) ",
          self.urlToken.localizedName);
    
    NSAssert( 0 != (UIDocumentStateClosed & self.document.documentState),
             @"Document must be in its closed state.");
    NSAssert( (UIDocumentStateClosed == self.document.documentState), 
             @"Document must be in its closed state exactly.");
    
    [self.operationQueue setSuspended: YES];
    [self.document openWithToken: self];
}
// ============
-(void)nqd_forceDownload
{
    [self.operationQueue setSuspended: YES];
    NSLog(@"-[PNDocToken nqd_forceDownload] (%@) ",
          self.urlToken.localizedName);

    NSAssert( [self.urlToken isKindOfClass: [PNCloudURLToken class]], @"Can't force download of a local url");
    
    PNCloudURLToken *cloudURLToken = (PNCloudURLToken*)(self.urlToken);
    
    // No need to suspend and resume the queue;
    // this is not a sequential action;
    // this kicks off a parallel action.
    [cloudURLToken forceDownload];
    [self.operationQueue setSuspended: NO];
}
// ============

-(void)nqd_saveDocumentForRenaming
{
    [self.operationQueue setSuspended: YES];
    NSLog(@"-[PNDocToken nqd_saveDocumentForOverwriting] (%@) ",
          self.urlToken.localizedName);
    
    NSAssert( (UIDocumentStateNormal == self.document.documentState),
             @"Document must be in its normal state.");
    // Can the document be in conflict?
    
    //[self logRegisteredObjects];
    
    // we're overwriting to rename, don't use the urlToken
    [self.document saveToURL: self.document.fileURL
            forSaveOperation: UIDocumentSaveForOverwriting
                       token: self];
}

-(void)nqd_saveDocumentForOverwriting
{
    [self.operationQueue setSuspended: YES];
    NSLog(@"-[PNDocToken nqd_saveDocumentForOverwriting] (%@) ",
          self.urlToken.localizedName);

    NSLog(@"self.document.documentState = %@",
          [self documentStateAsString]);
    
    enum UIDocumentState state = self.document.documentState;
    
    BOOL showStopper =
    (state & UIDocumentStateClosed) ||
    (state & UIDocumentStateInConflict) || //??
    (state & UIDocumentStateSavingError);
    
    // acceptable:
    //  UIDocumentStateNormal
    //  UIDocumentStateEditingDisabled
    
    NSAssert( !showStopper,
             @"Document must be in its normal state. Found %@",
             [self documentStateAsString]);
    // Can the document be in conflict?
    
    //[self logRegisteredObjects];
    
    // we're NOT overwriting to rename, so use the urlToken
    [self.document saveToURL: self.urlToken.localFilePackageURL
            forSaveOperation: UIDocumentSaveForOverwriting
                       token: self];
}
-(void)saveDocumentForOverwriting
{
    NSLog(@"-[PNDocToken saveDocumentForOverwriting]");
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_saveDocumentForOverwriting)];
    }[self closeQueue];
}
-(void)saveDocumentForRenaming
{
    
    /**
     This method expects that
     [1] the receiver's document's fileURL is still the original value; and
     [2] the receiver's url token has already been set to a new, distinct uuid (and name).
     
     See: http://stackoverflow.com/questions/8284137/what-is-the-proper-way-to-move-a-uidocument-to-a-new-location-on-the-file-system/8954496#8954496
     
     */
    [self logRegisteredObjects];
    
    NSLog(@"-[PNDocToken saveDocumentForRenaming]");
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_saveDocumentForRenaming)]; // uses self.document.fileURL
        [self enqueueSelector: @selector(nqd_closeDocument)];
        [self enqueueSelector: @selector(nqd_copyDocument)];
        
        if( [self.urlToken isKindOfClass: [PNCloudURLToken class]] ){
            [self enqueueSelector: @selector(nqd_makeUbiquitous)];
        }
        
        [self enqueueSelector: @selector(nqd_allocInitDocumentWithLocalURL)];
        // uses self.urlToken.localFilePackageURL
        // sets self.document.persistentStoreOptions = self.urlToken.persistentStoreOptions;
        
        [self enqueueSelector: @selector(nqd_openDocument)];
        [self enqueueSelector: @selector(nqd_saveDocumentForOverwriting)];
    }[self closeQueue];
    
}

-(void)saveDocumentForRenamingNEXT
{
    
    /**
     This method expects that 
     [1] the receiver's document's fileURL is still the original value; and
     [2] the receiver's url token has already been set to a new, distinct uuid (and name).

     See: http://stackoverflow.com/questions/8284137/what-is-the-proper-way-to-move-a-uidocument-to-a-new-location-on-the-file-system/8954496#8954496
     
     */
    [self logRegisteredObjects];
    
    NSLog(@"-[PNDocToken saveDocumentForRenaming]");
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_saveDocumentForOverwriting)];
        [self enqueueSelector: @selector(nqd_closeDocument)];
        [self enqueueSelector: @selector(nqd_copyDocument)];
        
        if( [self.urlToken isKindOfClass:[PNCloudURLToken class]] ){
            [self enqueueEstablishCloudDocument];
        }else{
            [self enqueueEstablishLocalDocument];
        }

    }[self closeQueue];
    
}

// ============
-(void)nqd_revert
{
    [self.operationQueue setSuspended: YES];
    
    NSLog(@"-[PNDocToken nqd_revert] (%@) ",
          self.urlToken.localizedName);
    
    UIDocumentState state = self.document.documentState;
    BOOL isClosed = state & UIDocumentStateClosed;
    if( isClosed ){
        NSAssert( !isClosed, 
                 @"Document must not be closed.");
        // Can the document be in conflict?
    }
    
    NSLog(@"-[PNDocToken revert]");
    //[self logRegisteredObjects];
    
    [self.operationQueue setSuspended: YES];
    [self.document revertToContentsOfURL: self.urlToken.localFilePackageURL
                                   token: self];
}

-(void)revert
{
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_revert)];
    }[self closeQueue];
}
// ============

-(void)nqd_closeDocument
{
    [self.operationQueue setSuspended: YES];
    
    NSLog(@"-[PNDocToken nqd_closeDocument] (%@) ",
          self.urlToken.localizedName);
    NSAssert( (UIDocumentStateNormal == self.document.documentState),
             @"Document must be in its normal state.");
    // Can the document be in conflict?
    
    NSLog(@"-[PNDocToken closeDocument]");
    //    NSLog(@"  self.document.fileURL     = \n%@", self.document.fileURL.absoluteString );
    //    NSLog(@"   self.urlToken.localFilePackageURL = \n%@", 
    //          self.urlToken.localFilePackageURL.absoluteString );
    
    NSManagedObjectContext *moc = self.document.managedObjectContext;
    [moc processPendingChanges];
    
    [self.document closeWithToken: self];
}
-(void)closeDocument
{
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_closeDocument)];
    }[self closeQueue];
}
-(void)closeDocumentInBackgroundTask
{
    // The receiver's marker or delegate 
    // may observe UIApplicationDidEnterBackgroundNotification
    // and call this method.
    
    NSAssert( (UIBackgroundTaskInvalid == m_bkgTaskID), 
             @"Background task already running." );
    
    UIApplication *sharedApp = [UIApplication sharedApplication];
    
    m_bkgTaskID =
    [sharedApp beginBackgroundTaskWithExpirationHandler: ^{
        
        NSLog( @"-[%@ closeAndFinalize] default expiration handler, start",
              NSStringFromClass([self class]) );
        NSLog( @"%@ launched expiration handler", 
              [self quickLook]);

    }];
    
    // The close operation  saves the document first.
    // See Apple's docs for -[UIDocument closeWithCompletionHandler:].
    
    // "You call this method to begin the sequence of method calls that 
    // saves a document safely and asynchronously. "
    
    // At the end of the enqueued operations, 
    // -becomeAvailable sends -acceptDocToken: to nil.
    self.creationDelegate = nil;
    
    [self openQueue];{
        
        [self enqueueSelector:@selector(nqd_closeDocument)];
        [self enqueueSelector:@selector(nqd_finalizeDocument)];
        
    }[self closeQueue];
    
}

// ============

-(void)nqd_finalizeDocument
{
    [self.operationQueue setSuspended: YES];
    
    NSLog(@"-[PNDocToken nqd_finalizeDocument] (%@) ",
          self.urlToken.localizedName);
 
    //    NSLog(@"  self.document.fileURL     = \n%@", self.document.fileURL.absoluteString );
    //    NSLog(@"   self.urlToken.localFilePackageURL = \n%@", 
    //          self.urlToken.localFilePackageURL.absoluteString );
    
    [self.operationQueue setSuspended: YES];
    m_document = nil;
    //[self.delegate didFinalizeDocToken: self];
    
    NSLog( @"uuid: %@ intended expiration handler, end", 
          self.urlToken.uuid);
    [[UIApplication sharedApplication] endBackgroundTask: self.bkgTaskID];
}
-(void)finalizeDocument
{
    [self openQueue];{
        [self enqueueSelector: @selector(nqd_finalizeDocument)];
    }[self closeQueue];
}

-(void)enqueueCreateLocalDocument
{
    
    NSAssert( self.creationDelegate, @"Bogus nil creation delegate.");
    
    NSAssert( [(NSObject*)(self.creationDelegate) conformsToProtocol:
               @protocol(PNDocumentCreating)],
               @"Bogus creation delegate doesn't conform to PNDocumentCreating.");
    
    enum PNURLDeploymentCase localURLDeploymentCase = [self.urlToken.localFilePackageURL deploymentCase];

    switch( localURLDeploymentCase ){
        case PNURLDeploymentCaseUuidDirSlashNoSuchFileOrDirectory:
        {
            
            // Create a LOCAL document:
            
            [self enqueueSelector:@selector(nqd_allocInitDocumentWithLocalURL)];
            [self enqueueSelector:@selector(nqd_saveDocumentLocallyForCreating)];
            [self enqueueSelector:@selector(nqd_configureNewDocument)];
            [self enqueueSelector:@selector(nqd_saveDocumentForOverwriting)];

            break;
        }
        case PNURLDeploymentCaseUuidDirSlashExistingFile:
        case PNURLDeploymentCaseUuidDirSlashExistingRWDir:
        {
            NSAssert( NO, @"When creating a doc, its file package must not exist already.");
            break;
        }
        case PNURLDeploymentCaseNoUuidDir: // build the uuid elsewhere, on the way here.
        {
            NSAssert( NO, @"When creating a doc, its uuid dir should have been created when the url token was created.");
            break;
        }
        case PNURLDeploymentCaseInvalid:        {
            NSAssert( NO, @"Oh, noz! nil url token?");
            break;
        }
        case PNURLDeploymentCaseDefective:
        {
            NSAssert( NO, @"Oh, noz! defective file package?");
            break;
        }
        default:
        {
            NSAssert( NO, @"Oh, noz! nonsense value of deploymentCase: %d ",
                     localURLDeploymentCase);
            break;
        }
    }

}
-(void)enqueueCreateCloudDocument
{
    
    NSAssert( self.creationDelegate, @"Bogus nil creation delegate.");
    
    NSAssert( [(NSObject*)(self.creationDelegate) conformsToProtocol:
               @protocol(PNDocumentCreating)],
             @"Bogus creation delegate doesn't conform to PNDocumentCreating.");
    
    enum PNURLDeploymentCase localURLDeploymentCase = [self.urlToken.localFilePackageURL deploymentCase];
    
    switch( localURLDeploymentCase ){
        case PNURLDeploymentCaseUuidDirSlashNoSuchFileOrDirectory:
        {
            // Create a CLOUD document:
            
            [self enqueueSelector:@selector(nqd_allocInitDocumentWithLocalURL)];
            [self enqueueSelector:@selector(nqd_saveDocumentLocallyForCreating)];
            [self enqueueSelector:@selector(nqd_configureNewDocument)];
            [self enqueueSelector:@selector(nqd_saveDocumentForOverwriting)];
            [self enqueueSelector:@selector(nqd_closeDocument)];
           
            //[self enqueueSelector:@selector(nqd_makeUbiquitous)];
            
            [self enqueueSelector:@selector(nqd_openDocument)];
            
            /**
             We may not need the complete Sadun sequence; 
             MultiDocumentPlusUUID does not.
             
            [self enqueueSelector:@selector(nqd_closeDocument)];
            
            [self enqueueSelector:@selector(nqd_makeUbiquitous)];
           
            // No, don't init from the cloud URL:
            //[self enqueueSelector:@selector(nqd_allocInitDocumentWithCloudURL)];
            
            // Instead, init from the local URL:
            [self enqueueSelector:@selector(nqd_allocInitDocumentWithLocalURL)];
            // The persistent store options determine the cloud URL.
            
            [self enqueueSelector:@selector(nqd_openDocument)];
*/
            break;
        }
        case PNURLDeploymentCaseUuidDirSlashExistingFile:
        case PNURLDeploymentCaseUuidDirSlashExistingRWDir:
        {
            NSAssert( NO, @"When creating a doc, its file package must not exist already.");
            break;
        }
        case PNURLDeploymentCaseNoUuidDir: // build the uuid elsewhere, on the way here.
        {
            NSAssert( NO, @"When creating a doc, its uuid dir should have been created when the url token was created.");
            break;
        }
        case PNURLDeploymentCaseInvalid:        {
            NSAssert( NO, @"Oh, noz! nil url token?");
            break;
        }
        case PNURLDeploymentCaseDefective:
        {
            NSAssert( NO, @"Oh, noz! defective file package?");
            break;
        }
        default:
        {
            NSAssert( NO, @"Oh, noz! nonsense value of deploymentCase: %d ",
                     localURLDeploymentCase);
            break;
        }
    }
    
}

-(void)enqueueEstablishLocalDocument
{
    
    [self enqueueSelector:@selector(nqd_allocInitDocumentWithLocalURL)];
    // uses self.urlToken.localFilePackageURL
    // sets self.document.persistentStoreOptions = self.urlToken.persistentStoreOptions;

    
    enum PNURLDeploymentCase thisCase = [self.urlToken.localFilePackageURL deploymentCase];
    
    switch( thisCase ){
        case PNURLDeploymentCaseUuidDirSlashNoSuchFileOrDirectory:
        case PNURLDeploymentCaseUuidDirSlashExistingFile:
        {
            [self enqueueSelector:@selector(nqd_saveDocumentLocallyForCreating)];
            
            if( nil != self.creationDelegate ){
                // We might be opening a document using the Sadun sequence
                // in the case of no existing local file.
                // In that case,
                // . we're initializing a place-holder URL for the cloud deamon
                // . we expect the cloud deamon to provide an existing shared document
                // . we don't give the delegate the opportunity
                //   to initialize its document's graph of objects.
                [self enqueueSelector:@selector(nqd_configureNewDocument)];
            }
            [self enqueueSelector:@selector(nqd_saveDocumentForOverwriting)];
            break;
        }
        case PNURLDeploymentCaseUuidDirSlashExistingRWDir:
        {
            
            [self enqueueSelector:@selector(nqd_openDocument)];
            [self enqueueSelector:@selector(nqd_saveDocumentForOverwriting)];
            
            break;
        }
        case PNURLDeploymentCaseInvalid:
        case PNURLDeploymentCaseNoUuidDir: // build the uuid elsewhere, on the way here.
        case PNURLDeploymentCaseDefective:
        default:
        {
            NSAssert( NO, @"Invalid url: %@", [self.urlToken.localFilePackageURL path] );
            break; // defensive, never happens. GLITCH comment?
        }
    }
    
}

-(void)enqueueEstablishCloudDocument
{
    
    [self enqueueSelector:@selector(nqd_allocInitDocumentWithLocalURL)];
    // uses self.urlToken.localFilePackageURL
    // sets self.document.persistentStoreOptions = self.urlToken.persistentStoreOptions;

    
    enum PNURLDeploymentCase localURLDeploymentCase = [self.urlToken.localFilePackageURL deploymentCase];
    
    switch( localURLDeploymentCase ){
        case PNURLDeploymentCaseUuidDirSlashNoSuchFileOrDirectory:
        {
            [self enqueueSelector:@selector(nqd_saveDocumentLocallyForCreating)];
            
            [self enqueueSelector:@selector(nqd_closeDocument)];
            
            [self enqueueSelector: @selector(nqd_makeUbiquitous)];
            
            // Once we close a document, we should not re-open it.
            // Make a new instance.
            // See: http://stackoverflow.com/questions/9563949/uidocument-couldnt-close-again
            // "It seems you can only close a UIManagedDocument once per instantiation."
            // "... the new instance of the document will have a new managed object context ..."
            //
            
            [self enqueueSelector:@selector(nqd_allocInitDocumentWithLocalURL)];
            // uses self.urlToken.localFilePackageURL
            // sets self.document.persistentStoreOptions = self.urlToken.persistentStoreOptions;

            
            // NOW we can open the new instance.
            [self enqueueSelector:@selector(nqd_openDocument)];
            
            break;
        }
        case PNURLDeploymentCaseUuidDirSlashExistingRWDir:
        {
            
            [self enqueueSelector:@selector(nqd_openDocument)];
            
            break;
        }
        case PNURLDeploymentCaseInvalid:
        case PNURLDeploymentCaseNoUuidDir: // build the uuid elsewhere, on the way here.
        case PNURLDeploymentCaseDefective:
        default:
        {
            NSAssert( NO, @"Invalid url: %@", [self.urlToken.localFilePackageURL path] );
            break; // defensive, never happens.
        }
    }
    
}

@end

@implementation PNDocToken (AggregateFileOperations)
#pragma mark Grouped Operations


/**
 
 PNCodingModel provides a switch between two ways of coding aggregated file operations. The "folded" approach follows examples by Sadun, Warren, et al. It's straightforward, but it doesn't scale well with greater complexity of aggregation. The "pipelined" approach follows the Automator model. It requires more infrastructure, but offers greater flexibility and yields far more readable code.
 
 I strongly prefer the pipelined approach. It works for the sandbox-only case. And it seems to work on the iPhone for the iCloud case for creating new iCloud documents and opening existing ones, the pipelined close-open operation. But, the same code fails on the iPad; for now (circa late April 2013), the iPad wants the folded approach for creating and opening documents. We'll need more clues from Apple to figure this out. As for the rest of the pipelined file operations, they hardly matter until iCloud syncing of UIManagedDocument proves reliable.
 
 In the meantime, I've thrown in PNCodingModel as a way to switch between the two coding models. This release is, after all, a test bench for iCloud syncing of UIManagedDocument.
 
 See for example: http://appleinsider.com/articles/13/03/29/apples-icloud-disparaged-over-core-data-sync-problems
 
 "Developer Steve Streza wrote, "the reality of syncing data is that it’s tough..."

 */
typedef enum PNCodingModel
{
	PNCodingModelInvalid = 0, /**< Programmer asleep at the switch */
	PNCodingModelFolded, /**< The Origami model */
	PNCodingModelPipelined /**< American craftsmen say that a plumber only has to know two things:
                            [1] Used resources flow downhill, and
                            [2] payday's Friday. */
} PNCodingModel;


static enum PNCodingModel codingModelUnderTest = PNCodingModelFolded;

-(enum PNCodingModel)codingModel
{
    enum PNCodingModel result = PNCodingModelInvalid;
    
    {
        switch( codingModelUnderTest ){
            case PNCodingModelFolded:
            {
                result = PNCodingModelFolded;
                break;
            }
            case PNCodingModelPipelined:
            {
                UIDevice *device = [UIDevice currentDevice];
                NSString *model = device.model;
                NSRange range = [model rangeOfString:@"iPhone"];
                BOOL isPhone = ( NSNotFound != range.location );
                
                if( isPhone ){
                    result = PNCodingModelPipelined;
                }else{
                    // With PNCodingModelPipelined on iPad,
                    // the NSMetadataQuery result count is always 0.
                    result = PNCodingModelFolded;
                }
                break;
            }
            default:
            {
                result = PNCodingModelInvalid;
                break;
            }
        }
    }
    return result;
}
-(void)createAndSaveDocument
{
    switch( codingModelUnderTest ){
        case PNCodingModelFolded:
        {
            [self createAndSaveDocumentFolded];
            // See PNDocToken+Managing
            
            break;
        }
        case PNCodingModelPipelined:
        {
            [self createAndSaveDocumentPipelined];
            break;
        }
        default:
        {
            NSAssert( NO, @" Programmer asleep at the switch");
            break;
        }
    }
}
/*
 See Erica Sadun,
 The iOS 5 Developer's Cookbook: The Additional Recipes: Additional Recipes
 Recipe: UIManagedDocument and Core Data
 Establish the Document
 */
-(void)establishDocument
{
    switch( codingModelUnderTest ){
            
        case PNCodingModelFolded:
        {
            [self establishDocumentFolded];
            // See PNDocToken+Managing
            
            break;
        }
        case PNCodingModelPipelined:
        {
            [self establishDocumentPipelined];
            break;
        }
        default:
        {
            NSAssert( NO, @" Programmer asleep at the switch");
            break;
        }
    }
}

-(void)createAndSaveDocumentPipelined
{
    [self openQueue];{
        
        if( [self.urlToken isKindOfClass:[PNCloudURLToken class]] ){
            [self enqueueCreateCloudDocument];
        }else{
            [self enqueueCreateLocalDocument];
        }
        
    }[self closeQueue];
}
-(void)establishDocumentPipelined
{
    [self openQueue];{
        
        if( [self.urlToken isKindOfClass:[PNCloudURLToken class]] ){
            [self enqueueEstablishCloudDocument];
        }else{
            [self enqueueEstablishLocalDocument];
        }
        
    }[self closeQueue];}
@end
